@@ -1,4 +1,5 @@
 from abc import abstractmethod, ABC
+from copy import copy
 from typing import Mapping, cast, Any
 from algosdk.future.transaction import StateSchema
 from pyteal import (
@@ -22,6 +23,25 @@ from pyteal import (
 from beaker.consts import MAX_GLOBAL_STATE, MAX_LOCAL_STATE
 
 
+def get_default_for_type(stack_type, default):
+    if default is not None:
+        return default
+
+    if stack_type == TealType.bytes:
+        return Bytes("")
+    else:
+        return Int(0)
+
+
+def stack_type_to_string(st: TealType):
+    if st == TealType.uint64:
+        return "uint64"
+    if st == TealType.bytes:
+        return "bytes"
+    else:
+        raise Exception("Only uint64 and bytes supported")
+
+
 class StateValue(Expr):
     def __init__(
         self,
@@ -62,22 +82,34 @@ class StateValue(Expr):
         """returns the string held by the key Bytes object"""
         return cast(Bytes, self.key).byte_str.replace('"', "")
 
-    @abstractmethod
+    def increment(self, cnt: Expr = Int(1)) -> Expr:
+        """helper to increment a counter"""
+        check_is_int(self)
+        check_not_static(self)
+
+        return self.set(self.get() + cnt)
+
+    def decrement(self, cnt: Expr = Int(1)) -> Expr:
+        """helper to decrement a counter"""
+        check_is_int(self)
+        check_not_static(self)
+
+        return self.set(self.get() - cnt)
+
     def set_default(self) -> Expr:
         """sets the default value if one is provided, if none provided sets the zero value for its type"""
 
+        return self.set(get_default_for_type(self.stack_type, self.default))
+
+    def is_default(self) -> Expr:
+        """checks to see if the value set equals the default value"""
+        default = get_default_for_type(self.stack_type, self.default)
+        return self.get() == default
+
     @abstractmethod
     def set(self, val: Expr) -> Expr:
         """sets the value to the argument passed"""
 
-    @abstractmethod
-    def increment(self, cnt: Expr = Int(1)) -> Expr:
-        """helper to increment a counter"""
-
-    @abstractmethod
-    def decrement(self, cnt: Expr = Int(1)) -> Expr:
-        """helper to decrement a counter"""
-
     @abstractmethod
     def get(self) -> Expr:
         """gets the value stored for this state value"""
@@ -98,10 +130,6 @@ class StateValue(Expr):
     def delete(self) -> Expr:
         """deletes the key from state, if the value is static it will be a compile time error"""
 
-    @abstractmethod
-    def is_default(self) -> Expr:
-        """checks to see if the value set equals the default value"""
-
 
 class DynamicStateValue(ABC):
     def __init__(
@@ -129,24 +157,11 @@ class ApplicationStateValue(StateValue):
     def __str__(self) -> str:
         return f"ApplicationStateValue {self.key}"
 
-    def set_default(self) -> Expr:
-        if self.key is None:
-            raise Exception("No Key defined for application state value")
-
-        if self.default:
-            return App.globalPut(self.key, self.default)
-
-        if self.stack_type == TealType.uint64:
-            return App.globalPut(self.key, Int(0))
-        else:
-            return App.globalPut(self.key, Bytes(""))
-
     def set(self, val: Expr) -> Expr:
-        if self.key is None:
-            raise Exception("No Key defined for application state value")
+        check_match_type(self, val)
 
-        if val.type_of() != self.stack_type and val.type_of() != TealType.anytype:
-            raise TealTypeError(val.type_of(), self.stack_type)
+        if self.key is None:
+            raise TealInputError(f"ApplicationStateValue {self} has no key defined")
 
         if self.static:
             return Seq(
@@ -157,66 +172,40 @@ class ApplicationStateValue(StateValue):
 
         return App.globalPut(self.key, val)
 
-    def increment(self, cnt: Expr = Int(1)) -> Expr:
-        if self.key is None:
-            raise Exception("No Key defined for application state value")
-
-        if self.stack_type != TealType.uint64:
-            raise TealInputError("Only uint64 types can be incremented")
-
-        if self.static:
-            raise TealInputError("Cannot increment a static value")
-
-        return self.set(self.get() + cnt)
-
-    def decrement(self, cnt: Expr = Int(1)) -> Expr:
-        if self.key is None:
-            raise Exception("No Key defined for application state value")
-
-        if self.stack_type != TealType.uint64:
-            raise TealInputError("Only uint64 types can be decremented")
-
-        if self.static:
-            raise TealInputError("Cannot decrement a static value")
-
-        return self.set(self.get() - cnt)
-
     def get(self) -> Expr:
         if self.key is None:
-            raise Exception("No Key defined for application state value")
+            raise TealInputError(f"ApplicationStateValue {self} has no key defined")
 
         return App.globalGet(self.key)
 
     def get_maybe(self) -> MaybeValue:
         if self.key is None:
-            raise Exception("No Key defined for application state value")
+            raise TealInputError(f"ApplicationStateValue {self} has no key defined")
 
         return App.globalGetEx(Int(0), self.key)
 
     def get_must(self) -> Expr:
         if self.key is None:
-            raise Exception("No Key defined for application state value")
+            raise TealInputError(f"ApplicationStateValue {self} has no key defined")
 
         return Seq(val := self.get_maybe(), Assert(val.hasValue()), val.value())
 
     def get_else(self, val: Expr) -> Expr:
+        check_match_type(self, val)
+
         if self.key is None:
-            raise Exception("No Key defined for application state value")
+            raise TealInputError(f"ApplicationStateValue {self} has no key defined")
 
         return If((v := App.globalGetEx(Int(0), self.key)).hasValue(), v.value(), val)
 
     def delete(self) -> Expr:
+        check_not_static(self)
+
         if self.key is None:
-            raise Exception("No Key defined for application state value")
+            raise TealInputError(f"ApplicationStateValue {self} has no key defined")
 
-        if self.static:
-            raise TealInputError("Cannot delete static global param")
         return App.globalDel(self.key)
 
-    def is_default(self) -> Expr:
-
-        return self.get() == self.default
-
 
 class DynamicApplicationStateValue(DynamicStateValue):
     def __init__(
@@ -249,101 +238,83 @@ class DynamicApplicationStateValue(DynamicStateValue):
 
 
 class AccountStateValue(StateValue):
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.acct = Txn.sender()
+
     def __str__(self) -> str:
-        return f"AccountStateValue {self.key}"
+        return f"AccountStateValue {self.acct} {self.key}"
+
+    def set(self, val: Expr) -> Expr:
+        check_match_type(self, val)
 
-    def set_default(self, acct: Expr = Txn.sender()) -> Expr:
         if self.key is None:
-            raise Exception("No Key defined for account state value")
+            raise TealInputError(f"AccountStateValue {self} has no key defined")
 
-        if self.default is not None:
-            return App.localPut(acct, self.key, self.default)
-
-        if self.stack_type == TealType.uint64:
-            return App.localPut(acct, self.key, Int(0))
-        else:
-            return App.localPut(acct, self.key, Bytes(""))
-
-    def set(self, val: Expr, acct: Expr = Txn.sender()) -> Expr:
-        if self.key is None:
-            raise Exception("No Key defined for account state value")
-
-        if val.type_of() != self.stack_type and val.type_of() != TealType.anytype:
-            raise TealTypeError(val.type_of(), self.stack_type)
+        if self.acct is None:
+            raise TealInputError(f"AccountStateValue {self} has no account defined")
 
         if self.static:
             return Seq(
-                v := self.get_maybe(acct),
+                v := self.get_maybe(),
                 Assert(Not(v.hasValue())),
-                App.localPut(acct, self.key, val),
+                App.localPut(self.acct, self.key, val),
             )
 
-        return App.localPut(acct, self.key, val)
+        return App.localPut(self.acct, self.key, val)
 
-    def get(self, acct: Expr = Txn.sender()) -> Expr:
+    def get(self) -> Expr:
         if self.key is None:
-            raise Exception("No Key defined for account state value")
+            raise TealInputError(f"AccountStateValue {self} has no key defined")
 
-        return App.localGet(acct, self.key)
+        if self.acct is None:
+            raise TealInputError(f"AccountStateValue {self} has no account defined")
 
-    def get_maybe(self, acct: Expr = Txn.sender()) -> MaybeValue:
+        return App.localGet(self.acct, self.key)
+
+    def get_maybe(self) -> MaybeValue:
         if self.key is None:
-            raise Exception("No Key defined for account state value")
+            raise TealInputError(f"AccountStateValue {self} has no key defined")
 
-        return App.localGetEx(acct, Int(0), self.key)
+        if self.acct is None:
+            raise TealInputError(f"AccountStateValue {self} has no account defined")
 
-    def get_must(self, acct: Expr = Txn.sender()) -> Expr:
+        return App.localGetEx(self.acct, Int(0), self.key)
+
+    def get_must(self) -> Expr:
         if self.key is None:
-            raise Exception("No Key defined for account state value")
+            raise TealInputError(f"AccountStateValue {self} has no key defined")
+        if self.acct is None:
+            raise TealInputError(f"AccountStateValue {self} has no account defined")
 
-        return Seq(val := self.get_maybe(acct), Assert(val.hasValue()), val.value())
+        return Seq(val := self.get_maybe(), Assert(val.hasValue()), val.value())
+
+    def get_else(self, val: Expr) -> Expr:
+        check_match_type(self, val)
 
-    def get_else(self, val: Expr, acct: Expr = Txn.sender()) -> Expr:
         if self.key is None:
-            raise Exception("No Key defined for account state value")
-
-        if val.type_of() != self.stack_type:
-            raise TealTypeError(val.type_of(), self.stack_type)
+            raise TealInputError(f"AccountStateValue {self} has no key defined")
+        if self.acct is None:
+            raise TealInputError(f"AccountStateValue {self} has no account defined")
 
         return If(
-            (v := App.localGetEx(acct, Int(0), self.key)).hasValue(), v.value(), val
+            (v := App.localGetEx(self.acct, Int(0), self.key)).hasValue(),
+            v.value(),
+            val,
         )
 
-    def delete(self, acct: Expr = Txn.sender()) -> Expr:
+    def delete(self) -> Expr:
         if self.key is None:
-            raise Exception("No Key defined for account state value")
+            raise TealInputError(f"AccountStateValue {self} has no key defined")
+        if self.acct is None:
+            raise TealInputError(f"AccountStateValue {self} has no account defined")
 
-        return App.localDel(acct, self.key)
+        return App.localDel(self.acct, self.key)
 
-    def increment(self, cnt: Expr = Int(1), acct: Expr = Txn.sender()) -> Expr:
-        if self.key is None:
-            raise Exception("No Key defined for account state value")
-
-        if self.stack_type != TealType.uint64:
-            raise TealInputError("Only uint64 types can be incremented")
-
-        if self.static:
-            raise TealInputError("Cannot increment a static value")
-
-        return self.set(self.get() + cnt, acct=acct)
-
-    def decrement(self, cnt: Expr = Int(1), acct: Expr = Txn.sender()) -> Expr:
-        if self.key is None:
-            raise Exception("No Key defined for account state value")
-
-        if self.stack_type != TealType.uint64:
-            raise TealInputError("Only uint64 types can be decremented")
-
-        if self.static:
-            raise TealInputError("Cannot decrement a static value")
-
-        return self.set(self.get() - cnt, acct=acct)
-
-    def is_default(self, acct: Expr = Txn.sender()) -> Expr:
-        if self.key is None:
-            raise Exception("No Key defined for account state value")
-
-        return self.get(acct) == self.default
+    def __getitem__(self, acct: Expr) -> "AccountStateValue":
+        asv = copy(self)
+        asv.acct = acct
+        return asv
 
 
 class DynamicAccountStateValue(DynamicStateValue):
@@ -354,7 +325,6 @@ class DynamicAccountStateValue(DynamicStateValue):
         key_gen: SubroutineFnWrapper = None,
         descr: str = None,
     ):
-
         super().__init__(stack_type, max_keys, key_gen, descr)
 
         if max_keys <= 0 or max_keys > MAX_LOCAL_STATE:
@@ -373,13 +343,20 @@ class DynamicAccountStateValue(DynamicStateValue):
         return AccountStateValue(stack_type=self.stack_type, key=cast(Expr, key))
 
 
-def stack_type_to_string(st: TealType):
-    if st == TealType.uint64:
-        return "uint64"
-    if st == TealType.bytes:
-        return "bytes"
-    else:
-        raise Exception("Only uint64 and bytes supported")
+def check_not_static(sv: StateValue):
+    if sv.static:
+        raise TealInputError(f"StateValue {sv} is static")
+
+
+def check_is_int(sv: StateValue):
+    if sv.stack_type != TealType.uint64:
+        raise TealInputError(f"StateValue {sv} is not integer type")
+
+
+def check_match_type(sv: StateValue, val: Expr):
+    in_type = val.type_of()
+    if in_type != sv.stack_type and in_type != TealType.anytype:
+        raise TealTypeError(in_type, sv.stack_type)
 
 
 class State:
@@ -421,11 +398,19 @@ class State:
         """Convert the state to a dict for encoding"""
         return {
             "declared": {
-                k: {"type": stack_type_to_string(v.stack_type), "key": v.str_key()}
+                k: {
+                    "type": stack_type_to_string(v.stack_type),
+                    "key": v.str_key(),
+                    "descr": v.descr,
+                }
                 for k, v in self.declared_vals.items()
             },
             "dynamic": {
-                k: {"type": stack_type_to_string(v.stack_type), "max-keys": v.max_keys}
+                k: {
+                    "type": stack_type_to_string(v.stack_type),
+                    "max-keys": v.max_keys,
+                    "descr": v.descr,
+                }
                 for k, v in self.dynamic_vals.items()
             },
         }
@@ -473,7 +458,7 @@ class AccountState(State):
         """Generate expression from state values to initialize a default value"""
         return Seq(
             *[
-                cast(AccountStateValue, v).set_default(acct)
+                cast(AccountStateValue, v)[acct].set_default()
                 for v in self.declared_vals.values()
                 if not v.static or (v.static and v.default is not None)
             ]
